"""
TABI - UCSC iGEM 2023

Module: CodonAnalyzer

Description: Classes to calculate codon usage statistics based on a collection of CDS from an organism

Author: Tyler Gaw, James Larbalestier
"""

from ._CodonAnalyzer import *
